@@ -1,12 +1,16 @@
-#!/usr/bin/python2
+#!/usr/bin/env python22
 # Terminator by Chris Jones <cmsj@tenshu.net>
 # GPL v2 only
 """terminator.py - class for the master Terminator singleton"""
 
 import copy
 import os
-from gi.repository import Gtk, Gdk
+import gi
+gi.require_version('Vte', '2.91')
+from gi.repository import Gtk, Gdk, Vte, GdkX11
+from gi.repository.GLib import GError
 
+import borg
 from borg import Borg
 from config import Config
 from keybindings import Keybindings
@@ -39,7 +43,7 @@ class Terminator(Borg):
     groups = None
     config = None
     keybindings = None
-    style_provider = None
+    style_providers = None
     last_focused_term = None
 
     origcwd = None
@@ -53,10 +57,14 @@ class Terminator(Borg):
     doing_layout = None
     layoutname = None
     last_active_window = None
+    prelayout_windows = None
 
     groupsend = None
     groupsend_type = {'all':0, 'group':1, 'off':2}
 
+    cur_gtk_theme_name = None
+    gtk_settings = None
+
     def __init__(self):
         """Class initialiser"""
 
@@ -81,12 +89,21 @@ class Terminator(Borg):
         if not self.keybindings:
             self.keybindings = Keybindings()
             self.keybindings.configure(self.config['keybindings'])
+        if not self.style_providers:
+            self.style_providers = []
         if not self.doing_layout:
             self.doing_layout = False
         if not self.pid_cwd:
             self.pid_cwd = get_pid_cwd()
         if self.gnome_client is None:
             self.attempt_gnome_client()
+        self.connect_signals()
+
+    def connect_signals(self):
+        """Connect all the gtk signals"""
+        self.gtk_settings=Gtk.Settings().get_default()
+        self.gtk_settings.connect('notify::gtk-theme-name', self.on_gtk_theme_name_notify)
+        self.cur_gtk_theme_name = self.gtk_settings.get_property('gtk-theme-name')
 
     def set_origcwd(self, cwd):
         """Store the original cwd our process inherits"""
@@ -201,6 +218,15 @@ class Terminator(Borg):
                 return terminal
         return None
 
+    def find_window_by_uuid(self, uuid):
+        """Search our terminals for one matching the supplied UUID"""
+        dbg('searching self.terminals for: %s' % uuid)
+        for window in self.windows:
+            dbg('checking: %s (%s)' % (window.uuid.urn, window))
+            if window.uuid.urn == uuid:
+                return window
+        return None
+
     def new_window(self, cwd=None, profile=None):
         """Create a window with a Terminal in it"""
         maker = Factory()
@@ -222,6 +248,8 @@ class Terminator(Borg):
         objects = {}
 
         self.doing_layout = True
+        self.last_active_window = None
+        self.prelayout_windows = self.windows[:]
 
         layout = copy.deepcopy(self.config.layout_get_config(layoutname))
         if not layout:
@@ -361,30 +389,171 @@ class Terminator(Borg):
                     term = self.find_terminal_by_uuid(window_last_active_term_mapping[window].urn)
                     term.ensure_visible_and_focussed()
 
+        # Build list of new windows using prelayout list
+        new_win_list = []
         for window in self.windows:
-            if window.uuid == self.last_active_window:
+            if window not in self.prelayout_windows:
+                new_win_list.append(window)
+        
+        # Make sure all new windows get bumped to the top
+        for window in new_win_list:
+            window.show()
+            window.grab_focus()
+            try:
+                t = GdkX11.x11_get_server_time(window.get_window())
+            except AttributeError:
+                t = 0
+            window.get_window().focus(t)
+
+        # Awful workaround to be sure that the last focused window is actually the one focused.
+        # Don't ask, don't tell policy on this. Even this is not 100%
+        if self.last_active_window:
+            window = self.find_window_by_uuid(self.last_active_window.urn)
+            count = 0
+            while count < 1000 and Gtk.events_pending():
+                count += 1
+                Gtk.main_iteration_do(False)
                 window.show()
+                window.grab_focus()
+                try:
+                    t = GdkX11.x11_get_server_time(window.get_window())
+                except AttributeError:
+                    t = 0
+                window.get_window().focus(t)
+
+        self.prelayout_windows = None
+
+    def on_gtk_theme_name_notify(self, settings, prop):
+        """Reconfigure if the gtk theme name changes"""
+        new_gtk_theme_name = settings.get_property(prop.name)
+        if new_gtk_theme_name != self.cur_gtk_theme_name:
+            self.cur_gtk_theme_name = new_gtk_theme_name
+            self.reconfigure()
 
     def reconfigure(self):
         """Update configuration for the whole application"""
 
-        if self.style_provider is not None:
-            Gtk.StyleContext.remove_provider_for_screen(
-                Gdk.Screen.get_default(),
-                self.style_provider)
-            self.style_provider = None
-        if self.config['handle_size'] in xrange(0, 6):
-            css = """
-                GtkPaned {
-                    -GtkPaned-handle-size: %s
-                }
+        if self.style_providers != []:
+            for style_provider in self.style_providers:
+                Gtk.StyleContext.remove_provider_for_screen(
+                    Gdk.Screen.get_default(),
+                    style_provider)
+        self.style_providers = []
+
+        # Force the window background to be transparent for newer versions of
+        # GTK3. We then have to fix all the widget backgrounds because the
+        # widgets theming may not render it's own background.
+        css = """
+            .terminator-terminal-window {
+                background-color: alpha(@theme_bg_color,0); }
+
+            .terminator-terminal-window .notebook.header,
+            .terminator-terminal-window notebook header {
+                background-color: @theme_bg_color; }
+
+            .terminator-terminal-window .pane-separator {
+                background-color: @theme_bg_color; }
+
+            .terminator-terminal-window .terminator-terminal-searchbar {
+                background-color: @theme_bg_color; }
+            """
+
+        # Fix several themes that put a borders, corners, or backgrounds around
+        # viewports, making the titlebar look bad.
+        css += """
+            .terminator-terminal-window GtkViewport,
+            .terminator-terminal-window viewport {
+                border-width: 0px;
+                border-radius: 0px;
+                background-color: transparent; }
+            """
+
+        # Add per profile snippets for setting the background of the HBox
+        template = """
+            .terminator-profile-%s {
+                background-color: alpha(%s, %s); }
+            """
+        profiles = self.config.base.profiles
+        for profile in profiles.keys():
+            if profiles[profile]['use_theme_colors']:
+                # Create a dummy window/vte and realise it so it has correct
+                # values to read from
+                tmp_win = Gtk.Window()
+                tmp_vte = Vte.Terminal()
+                tmp_win.add(tmp_vte)
+                tmp_win.realize()
+                bgcolor = tmp_vte.get_style_context().get_background_color(Gtk.StateType.NORMAL)
+                bgcolor = "#{0:02x}{1:02x}{2:02x}".format(int(bgcolor.red  * 255),
+                                                          int(bgcolor.green * 255),
+                                                          int(bgcolor.blue * 255))
+                tmp_win.remove(tmp_vte)
+                del(tmp_vte)
+                del(tmp_win)
+            else:
+                bgcolor = Gdk.RGBA()
+                bgcolor = profiles[profile]['background_color']
+            if profiles[profile]['background_type'] == 'transparent':
+                bgalpha = profiles[profile]['background_darkness']
+            else:
+                bgalpha = "1"
+
+            munged_profile = "".join([c if c.isalnum() else "-" for c in profile])
+            css += template % (munged_profile, bgcolor, bgalpha)
+
+        style_provider = Gtk.CssProvider()
+        style_provider.load_from_data(css)
+        self.style_providers.append(style_provider)
+
+        # Attempt to load some theme specific stylistic tweaks for appearances
+        usr_theme_dir = os.path.expanduser('~/.local/share/themes')
+        (head, _tail) = os.path.split(borg.__file__)
+        app_theme_dir = os.path.join(head, 'themes')
+
+        theme_name = self.gtk_settings.get_property('gtk-theme-name')
+
+        theme_part_list = ['terminator.css']
+        if self.config['extra_styling']:    # checkbox_style - needs adding to prefs
+            theme_part_list.append('terminator_styling.css')
+        for theme_part_file in theme_part_list:
+            for theme_dir in [usr_theme_dir, app_theme_dir]:
+                path_to_theme_specific_css = os.path.join(theme_dir,
+                                                          theme_name,
+                                                          'gtk-3.0/apps',
+                                                          theme_part_file)
+                if os.path.isfile(path_to_theme_specific_css):
+                    style_provider = Gtk.CssProvider()
+                    style_provider.connect('parsing-error', self.on_css_parsing_error)
+                    try:
+                        style_provider.load_from_path(path_to_theme_specific_css)
+                    except GError:
+                        # Hmmm. Should we try to provide GTK version specific files here on failure?
+                        gtk_version_string = '.'.join([str(Gtk.get_major_version()),
+                                                       str(Gtk.get_minor_version()),
+                                                       str(Gtk.get_micro_version())])
+                        err('Error(s) loading css from %s into Gtk %s' % (path_to_theme_specific_css,
+                                                                          gtk_version_string))
+                    self.style_providers.append(style_provider)
+                    break
+
+        # Size the GtkPaned splitter handle size.
+        css = ""
+        if self.config['handle_size'] in xrange(0, 21):
+            css += """
+                .terminator-terminal-window GtkPaned,
+                .terminator-terminal-window paned {
+                    -GtkPaned-handle-size: %s; }
                 """ % self.config['handle_size']
-            self.style_provider = Gtk.CssProvider()
-            self.style_provider.load_from_data(css)
+        style_provider = Gtk.CssProvider()
+        style_provider.load_from_data(css)
+        self.style_providers.append(style_provider)
+
+        # Apply the providers, incrementing priority so they don't cancel out
+        # each other
+        for idx in xrange(0, len(self.style_providers)):
             Gtk.StyleContext.add_provider_for_screen(
-                Gdk.Screen.get_default(), 
-                self.style_provider,     
-                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
+                Gdk.Screen.get_default(),
+                self.style_providers[idx],
+                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION+idx)
 
         # Cause all the terminals to reconfigure
         for terminal in self.terminals:
@@ -400,6 +569,15 @@ class Terminator(Borg):
             if maker.isinstance(child, 'Notebook'):
                 child.configure()
 
+    def on_css_parsing_error(self, provider, section, error, user_data=None):
+        """Report CSS parsing issues"""
+        file_path = section.get_file().get_path()
+        line_no = section.get_end_line() +1
+        col_no = section.get_end_position() + 1
+        err('%s, at line %d, column %d, of file %s' % (error.message,
+                                                       line_no, col_no,
+                                                       file_path))
+
     def create_group(self, name):
         """Create a new group"""
         if name not in self.groups:
